/* 
 * smcell.c --
 *
 *	This file implements cells.
 *
 */

#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "smInt.h"


/*
 * Prototypes for procedures defined externally:
 */

int		SmNormalizeVector3D _ANSI_ARGS_ ((DOUBLE vin[], DOUBLE vout[], double epsilon));

void		SmFastNormalizeVector3D _ANSI_ARGS_ ((DOUBLE vin[], DOUBLE vout[]));

void		SmCrossVector3D _ANSI_ARGS_ ((DOUBLE a[], DOUBLE b[], DOUBLE c[]));

void		SmDisplayViewport _ANSI_ARGS_ ((Tk_Canvas canvas,
						Tk_Item *itemPtr, Display *dpy, Drawable dst,
						int x, int y, int width, int height));

void            SmViewportUpdate _ANSI_ARGS_((ClientData clientData));

void		SmSetClientData _ANSI_ARGS_ ((ClientData clientData));


/*
 * Prototypes for procedures defined in this file that are exported:
 */

int		Sm_CellInit _ANSI_ARGS_ ((Tcl_Interp *interp));

void		SmInsertPortIntoCell _ANSI_ARGS_ ((Cell *cell, PortItem *port));

void		SmRemovePortFromCell _ANSI_ARGS_ ((Cell *cell, PortItem *port));

void		SmRemoveObjectFromParent _ANSI_ARGS_ ((Object *obj));

Cell		*Sm_FindCell _ANSI_ARGS_ ((Tcl_Interp *interp, char *label));

Object		*Sm_FindObject _ANSI_ARGS_ ((Cell *cell, char *label));

void		SmRedrawCell _ANSI_ARGS_ ((Cell *cell));


/*
 * Prototypes for procedures defined in this file that are static:
 */

static int		HandleCommand _ANSI_ARGS_ ((ClientData clientData, Tcl_Interp *interp, int argc, char **argv));

static int		HandleCellCommand _ANSI_ARGS_ ((ClientData clientData, Tcl_Interp *interp, int argc, char **argv));

static void		DestroyCell _ANSI_ARGS_ ((ClientData clientData));

static void		DestroyCells _ANSI_ARGS_ ((ClientData clientData));

static int		HandleNewObject _ANSI_ARGS_ ((Cell *cell, int argc, char **argv));

static int		NewPolyObject _ANSI_ARGS_ ((Cell *cell, int argc, char **argv));

static int		NewCylinderObject _ANSI_ARGS_ ((Cell *cell, int argc, char **argv));

static int		NewDiskObject _ANSI_ARGS_ ((Cell *cell, int argc, char **argv));

static int		NewSphereObject _ANSI_ARGS_ ((Cell *cell, int argc, char **argv));

static int		ConfigureObject _ANSI_ARGS_ ((Cell *cell, Object *obj, int argc, char **argv, int flags));

static void		DeleteObject _ANSI_ARGS_ ((Object *obj));

static void		FreeObject _ANSI_ARGS_ ((Object *obj));

static int		ComputeVertexNormals _ANSI_ARGS_ ((Object *obj));

static int		ComputeModelTransform _ANSI_ARGS_ ((Object *obj, double epsilon));

static int		ParseVertexNormals _ANSI_ARGS_ ((Tcl_Interp *interp, Object *obj, char *value));

static CellHeader	*GetCellHeader _ANSI_ARGS_ ((Tcl_Interp *interp));

static void		InsertCell _ANSI_ARGS_ ((Tcl_Interp *interp, Cell *cell));

static void		SmComputeSurfaceNormal _ANSI_ARGS_ ((Vertex *v, Surface *s, DOUBLE n[3], double epsilon));


/*
 * Custom options for cells
 */


/*
 * Information used for parsing cell configuration specs:
 */

static Tk_ConfigSpec cellConfigSpecs[] = {

    {TK_CONFIG_DOUBLE, "-epsilon", "epsilon", "Epsilon",
       "1e-5", Tk_Offset(Cell, epsilon), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};


/*
 * Information used for parsing object configuration specs:
 */

extern Tk_ConfigSpec	objectPolygonConfigSpecs[];
extern Tk_ConfigSpec	objectCylinderConfigSpecs[];
extern Tk_ConfigSpec	objectDiskConfigSpecs[];
extern Tk_ConfigSpec	objectSphereConfigSpecs[];
extern Tk_ItemType	tkViewportType;



/*
 *--------------------------------------------------------------
 *
 * Sm_CellInit:
 *
 * initialize the Cell package.
 *
 *--------------------------------------------------------------
 */

int
Sm_CellInit(Tcl_Interp *interp)
{
    CellHeader *header;

    if ((header = (CellHeader *) malloc(sizeof(CellHeader)))) {
	header->cells = NULL;
	Tcl_InitHashTable(&header->hash, TCL_STRING_KEYS);
	Tcl_CreateCommand(interp, "cell", HandleCommand, (ClientData) header, (Tcl_CmdDeleteProc *) DestroyCells);
	Tk_CreateItemType(&tkViewportType);

	return TCL_OK;
    }
    else {
	Tcl_AppendResult(interp, "Cell package initialization failed: malloc returns NULL", (char *) NULL);
	return TCL_ERROR;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Sm_FindCell:
 *
 * locate a cell by its label.
 *
 *--------------------------------------------------------------
 */

Cell *
Sm_FindCell(interp, label)
     Tcl_Interp *interp;
     char *label;
{
    CellHeader *header;
    Tcl_HashEntry *e;

    if ((header = GetCellHeader(interp)) == NULL) return NULL;
    if ((e = Tcl_FindHashEntry(&header->hash, label)) == NULL) return NULL;
    return (Cell *) Tcl_GetHashValue(e);
}

/*
 *--------------------------------------------------------------
 *
 * SmInsertPortIntoCell:
 *
 * insert a viewport into the cell viewport list.
 *
 *--------------------------------------------------------------
 */

void
SmInsertPortIntoCell(cell, port)
     Cell *cell;
     PortItem *port;
{
    port->next = cell->ports;
    cell->ports = port;
}

/*
 *--------------------------------------------------------------
 *
 * SmRemovePortFromCell:
 *
 * remove a viewport from the cell viewport list.
 *
 *--------------------------------------------------------------
 */

void
SmRemovePortFromCell(cell, port)
     Cell *cell;
     PortItem *port;
{
    PortItem *p, *q;

    for (p = cell->ports, q = NULL; p; q = p, p = p->next) {
	if (p == port) {
	    if (q) {
		q->next = p->next;
	    }
	    else {
		cell->ports = p->next;
	    }
	    return;
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * GetCellHeader:
 *
 * return cell header pointer.
 *
 *--------------------------------------------------------------
 */

static CellHeader *
GetCellHeader(interp)
     Tcl_Interp *interp;
{
    Tcl_CmdInfo info;

    return Tcl_GetCommandInfo(interp, "cell", &info) ? (CellHeader *) info.clientData : NULL;
}

/*
 *--------------------------------------------------------------
 *
 * InsertCell:
 *
 * insert a new cell into list of cells.
 *
 *--------------------------------------------------------------
 */

static void
InsertCell(interp, cell)
     Tcl_Interp *interp;
     Cell *cell;
{
    Tcl_CmdInfo info;
    CellHeader *header;

    if (!(Tcl_GetCommandInfo(interp, "cell", &info))) {
	return;
    }

    header = (CellHeader *) info.clientData;
    if (header->cells) {
	cell->next = header->cells;
    }
    else {
	cell->next = NULL;
    }
    header->cells = cell;
}

/*
 *--------------------------------------------------------------
 *
 * FreeObject:
 *
 * release memory occupied by an object.
 *
 *--------------------------------------------------------------
 */

static void
FreeObject(obj)
     Object *obj;
{
    int i;
    Surface *s;
    Object *ch, *next;

    SmRemoveObjectFromParent(obj);

    if (obj->label) free((void *) obj->label);
    if (obj->v) free((void *) obj->v);
    if (obj->e) Tcl_DeleteHashEntry(obj->e);
    if (obj->color != NULL) Tk_FreeColor(obj->color);
    if (obj->bfcolor != NULL) Tk_FreeColor(obj->bfcolor);
    if (obj->quadric != NULL) gluDeleteQuadric(obj->quadric);
    if (obj->normals) free((void *) obj->normals);

    if (obj->s) {
	for (s = obj->s, i = 0; i < obj->ns; i++) {
	    free((void *) s[i].index);
	}
	free((void *) obj->s);
    }

    for (ch = obj->child; ch; ch = next) {
	next = ch->sibling;
	DeleteObject(ch);
    }

    free((void *) obj);
}

/*
 *--------------------------------------------------------------
 *
 * DeleteObject:
 *
 * delete an object.
 *
 *--------------------------------------------------------------
 */

static void
DeleteObject(obj)
     Object *obj;
{
    Object *p, *q;

    for (p = obj->cell->objects, q = NULL; p; q = p, p = p->next) {
	if (p == obj) {
	    if (q) {
		q->next = p->next;
	    } else {
		obj->cell->objects = p->next;
	    }
	    FreeObject(obj);
	    return;
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * Sm_FindObject:
 *
 * locate an object within a cell by its label.
 *
 *--------------------------------------------------------------
 */

Object *
Sm_FindObject(cell, label)
     Cell *cell;
     char *label;
{
    Tcl_HashEntry *e;

    if ((e = Tcl_FindHashEntry(&cell->hash, label)) == NULL) {
	return NULL;
    }
    return (Object *) Tcl_GetHashValue(e);
}

/*
 *--------------------------------------------------------------
 *
 * SmRemoveObjectFromParent:
 *
 * dissociate an object from its parent object.
 *
 *--------------------------------------------------------------
 */

void
SmRemoveObjectFromParent(obj)
     Object *obj;
{
    Object *p, *sib;

    if ((p = obj->parent)) {
	if (p->child == obj) {
	    p->child = obj->sibling;
	} else {
	    for (sib = p->child; sib; sib = sib->sibling) {
		if (sib->sibling == obj) {
		    sib->sibling = obj->sibling;
		    return;
		}
	    }
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * SmRedrawCell:
 *
 * redraw all viewports associated with the cell.
 *
 *--------------------------------------------------------------
 */

void
SmRedrawCell(cell)
     Cell *cell;
{
    PortItem *p;

    for (p = cell->ports; p; p = p->next) {
	p->redraw = 1;
	if (p->canvas) {
	    Tk_CanvasEventuallyRedraw(p->canvas, ((Tk_Item *) p)->x1, ((Tk_Item *) p)->y1, ((Tk_Item *) p)->x2, ((Tk_Item *) p)->y2);
	} else if (!p->update) {
	    p->update = 1;
	    Tk_DoWhenIdle(SmViewportUpdate, (ClientData) p);
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * HandleCommand:
 *
 * handle cell commands:
 *
 *	cell create <label>
 *	cell list
 *
 *--------------------------------------------------------------
 */

static int
HandleCommand(clientData, interp, argc, argv)
     ClientData clientData;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    char c;
    int new, length;
    Cell *cell;
    CellHeader *header;
    Tcl_HashEntry *e;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " command ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = *argv[1];
    length = strlen(argv[1]);
    header = GetCellHeader(interp);

    if ((c == 'c') && (strncmp(argv[1], "create", length) == 0) && (length >= 2)) {

	/* command 'create' */
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " create label\"", (char *) NULL);
	    return TCL_ERROR;
	}

	e = Tcl_CreateHashEntry(&header->hash, argv[2], &new);

	if (!new) {
	    Tcl_AppendResult(interp, "a cell with label \"", argv[2], "\" already exists.", (char *) NULL);
	    return TCL_ERROR;
	}

	if ((cell = (Cell *) malloc(sizeof(Cell))) == NULL) {
	    Tcl_DeleteHashEntry(e);
	    Tcl_AppendResult(interp, "failed to allocate memory for new cell.", (char *) NULL);
	    return TCL_ERROR;
	}

	if (Tk_ConfigureWidget(interp, Tk_MainWindow(interp), cellConfigSpecs, argc - 3, argv + 3,
			       (char *) cell, 0) != TCL_OK) {
	    Tcl_DeleteHashEntry(e);
	    free((void *) cell);
	    return TCL_ERROR;
	}

	if ((cell->label = (char *) malloc(strlen(argv[2]) + 10)) == NULL) {
	    Tcl_DeleteHashEntry(e);
	    free((void *) cell);
	    Tcl_AppendResult(interp, "failed to allocate memory for new cell (label).", (char *) NULL);
	    return TCL_ERROR;
	}

	(void) strcpy(cell->label, argv[2]);
	cell->objects = NULL;
	cell->ports = NULL;
	cell->interp = interp;
	InsertCell(interp, cell);

	Tcl_InitHashTable(&cell->hash, TCL_STRING_KEYS);
	Tcl_SetHashValue(e, (ClientData) cell);
	Tcl_CreateCommand(interp, argv[2], HandleCellCommand, (ClientData) cell, (Tcl_CmdDeleteProc *) DestroyCell);
    }
    else if ((c == 'l') && (strncmp(argv[1], "list", length) == 0) && (length >= 2)) {

	/* command 'list' */
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " list\"", (char *) NULL);
	    return TCL_ERROR;
	}

	for (cell = header->cells; cell; cell = cell->next) {
	    Tcl_AppendElement(interp, cell->label);
	}
    }
    else {
	Tcl_AppendResult(interp, "bad command \"", argv[1], "\": must be configure, create or list.",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * HandleCellCommand:
 *
 * handle commands associated with a cell:
 *
 *	<cmd> modelconfigure <label> [options...]
 *	<cmd> modelcget <label> option
 *	<cmd> create <type> <label> [options...]
 *	<cmd> delete <label>
 *	<cmd> destroy
 *
 *--------------------------------------------------------------
 */

static int
HandleCellCommand(clientData, interp, argc, argv)
     ClientData clientData;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    char c;
    int length;
    Object *obj;
    Tk_Window tkwin;
    CellHeader *header;
    Tk_ConfigSpec *specs;
    Tcl_HashEntry *e;

    Cell *cell = (Cell *) clientData;

    tkwin = Tk_MainWindow(interp);
    header = GetCellHeader(interp);

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " command ?args?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = *argv[1];
    length = strlen(argv[1]);

    if ((c == 'm') && (strncmp(argv[1], "modelconfigure", length) == 0) && (length >= 2)) {

	/* command 'modelconfigure' */
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " modelconfigure label ?options?\"",
			   (char *) NULL);
	    return TCL_ERROR;
	}
	    
	if ((obj = Sm_FindObject(cell, argv[2])) == NULL) {
	    Tcl_AppendResult(interp, "object \"", argv[2], "\" not found in cell \"", cell->label, "\".",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (obj->type == SM_TYPE_CYLINDER) {
	    specs = objectCylinderConfigSpecs;
	}
	else if (obj->type == SM_TYPE_DISK) {
	    specs = objectDiskConfigSpecs;
	}
	else if (obj->type == SM_TYPE_SPHERE) {
	    specs = objectSphereConfigSpecs;
	}
	else if (obj->type == SM_TYPE_POLYGON) {
	    specs = objectPolygonConfigSpecs;
	}
	switch(argc) {

	  case 3:
	    return Tk_ConfigureInfo(interp, tkwin, specs, (char *) obj, NULL, TK_CONFIG_ARGV_ONLY);
	
	  case 4:
	    return Tk_ConfigureInfo(interp, tkwin, specs, (char *) obj, argv[3], TK_CONFIG_ARGV_ONLY);

	  default:
	    if (ConfigureObject(cell, obj, argc-3, argv+3, TK_CONFIG_ARGV_ONLY) != TCL_OK) {
		return TCL_ERROR;
	    }
	    SmRedrawCell(cell);
	}
    }
    else if ((c == 'm') && (strncmp(argv[1], "modelcget", length) == 0) && (length >= 2)) {

	/* command 'modelcget' */
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " modelcget label option\"", (char *) NULL);
	    return TCL_ERROR;
	}
	    
	if ((obj = Sm_FindObject(cell, argv[2])) == NULL) {
	    Tcl_AppendResult(interp, "object \"", argv[2], "\" not found in cell \"", cell->label, "\".",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (obj->type == SM_TYPE_CYLINDER) {
	    specs = objectCylinderConfigSpecs;
	}
	else if (obj->type == SM_TYPE_DISK) {
	    specs = objectDiskConfigSpecs;
	}
	else if (obj->type == SM_TYPE_SPHERE) {
	    specs = objectSphereConfigSpecs;
	}
	else if (obj->type == SM_TYPE_POLYGON) {
	    specs = objectPolygonConfigSpecs;
	}
	return Tk_ConfigureValue(interp, tkwin, specs, (char *) obj, argv[3], 0);
    }
    else if ((c == 'c') && (strncmp(argv[1], "create", length) == 0) && (length >= 2)) {
	
	/* command 'create' */
	if  (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " create type label ?options?\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (HandleNewObject(cell, argc, argv) != TCL_OK) {
	    return TCL_ERROR;
	}
	SmRedrawCell(cell);
    }
    else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0) && (length >= 2)) {

	/* command 'cget' */
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " cget option\"", (char *) NULL);
	    return TCL_ERROR;
	}
	return Tk_ConfigureInfo(interp, Tk_MainWindow(interp), cellConfigSpecs, (char *) cell,
				argv[3], TK_CONFIG_ARGV_ONLY);
    }
    else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0) && (length >= 2)) {

	Tk_Window tkwin = Tk_MainWindow(interp);

	/* command 'configure' */
	if (argc < 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " configure ?options?\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}

	switch(argc) {

	  case 2:
	    return Tk_ConfigureInfo(interp, tkwin, cellConfigSpecs, (char *) cell, NULL, TK_CONFIG_ARGV_ONLY);

	  case 3:
	    return Tk_ConfigureInfo(interp, tkwin, cellConfigSpecs, (char *) cell, argv[3], TK_CONFIG_ARGV_ONLY);

	  default:
	    if (Tk_ConfigureWidget(interp, tkwin, cellConfigSpecs, argc - 2, argv + 2, (char *) cell,
				   TK_CONFIG_ARGV_ONLY) != TCL_OK) {
		return TCL_ERROR;
	    }

	    for (obj = cell->objects; obj; obj = obj->next) {
		obj->invalid = ComputeModelTransform(obj, cell->epsilon);
	    }
	    SmRedrawCell(cell);
	    return TCL_OK;
	}
    }
    else if ((c == 'd') && (strncmp(argv[1], "delete", length) == 0) && (length >= 2)) {

	Object *p, *q;

	/* command 'delete' */
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " delete label\"", (char *) NULL);
	    return TCL_ERROR;
	}
	for (p = cell->objects, q = NULL; p; q = p, p = p->next) {
	    if (!strcmp(p->label, argv[2])) {
		DeleteObject(p);
		SmRedrawCell(cell);
		return TCL_OK;
	    }
	}
	Tcl_AppendResult(interp, "object \"", argv[2], "\" not found in cell \"", cell->label, "\".", (char *) NULL);
	return TCL_ERROR;
    }
    else if ((c == 'l') && (strncmp(argv[1], "list", length) == 0) && (length >= 2)) {

	/* command 'list' */
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " list\"", (char *) NULL);
	    return TCL_ERROR;
	}

	for (obj = cell->objects; obj; obj = obj->next) {
	    Tcl_AppendElement(interp, obj->label);
	}
    }
    else if ((c == 'd') && (strncmp(argv[1], "destroy", length) == 0) && (length >= 2)) {

	/* command 'destroy' */
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " destroy\"", (char *) NULL);
	    return TCL_ERROR;
	}

	if ((e = Tcl_FindHashEntry(&header->hash, cell->label)) == NULL) {
	    Tcl_AppendResult(interp, "cell \"", cell->label, "\" does not exist.", (char *) NULL);
	    return TCL_ERROR;
	}
	Tcl_DeleteCommand(interp, cell->label);
    }
    else {
	Tcl_AppendResult(interp, "bad cell command \"", argv[1],
			 "\": must be cget, configure, create, delete, destroy, or list.", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DestroyCell:
 *
 * handle the destruction of cells.
 *
 *--------------------------------------------------------------
 */

static void
DestroyCell(clientData)
     ClientData clientData;
{
    Cell *current, *last;
    CellHeader *header;
    Tcl_HashEntry *e;
    Object *obj;
    PortItem *p, *q;

    Cell *cell = (Cell *) clientData;

    header = GetCellHeader(cell->interp);
    for (obj = cell->objects; obj;) {
	if (obj->parent) {
	    obj = obj->next;
	} else {
	    DeleteObject(obj);
	    obj = cell->objects;
	}
    }

    for (p = cell->ports; p; p = q) {
	q = p->next;
	p->cell = NULL;
	p->next = NULL;
	p->redraw = 1;
	if (p->canvas) {
	    Tk_CanvasEventuallyRedraw(p->canvas, ((Tk_Item *) p)->x1, ((Tk_Item *) p)->y1,
				      ((Tk_Item *) p)->x2, ((Tk_Item *) p)->y2);
	} else if (p->tkwin) {
	    if (!p->update) {
		Tk_DoWhenIdle(SmViewportUpdate, (ClientData) p);
		p->update = 1;
	    }
	}
    }

    for (current = header->cells, last = NULL; current; last = current, current = current->next) {
	if (current == cell) {
	    if (last) {
		last->next = current->next;
	    } else {
		header->cells = current->next;
	    }
	    break;
	}
    }
    e = Tcl_FindHashEntry(&header->hash, cell->label);
    Tcl_DeleteHashTable(&cell->hash);
    Tcl_DeleteHashEntry(e);
    free((void *) cell->label);

    free((void *) cell);
}

/*
 *--------------------------------------------------------------
 *
 * DestroyCells:
 *
 * destroy all cells..
 *
 *--------------------------------------------------------------
 */

static void
DestroyCells(clientData)
     ClientData clientData;
{
    CellHeader *header = (CellHeader *) clientData;

    while (header->cells) {
	DestroyCell((ClientData) header->cells);
    }
}

/*
 *--------------------------------------------------------------
 *
 * HandleNewObject:
 *
 * handle the creation of new objects.
 *
 *	syntax of the input is: <cmd> new <type> <label> ?options?
 *	valid object types are:
 *		'polygon'
 *		'cylinder'
 *		'disk'
 *		'sphere'
 *
 *--------------------------------------------------------------
 */

static int
HandleNewObject(cell, argc, argv)
     Cell *cell;
     int argc;
     char **argv;
{
    int length;
    char c;
    Tcl_Interp *interp = cell->interp;

    c = argv[2][0];
    length = strlen(argv[2]);

    if ((c == 'p') && (strncmp(argv[2], "polygon", length) == 0) && (length >= 2)) {
	return NewPolyObject(cell, argc-3, argv+3);
    }
    else if ((c == 'c') && (strncmp(argv[2], "cylinder", length) == 0) && (length >= 2)) {
	return NewCylinderObject(cell, argc-3, argv+3);
    }
    else if ((c == 'd') && (strncmp(argv[2], "disk", length) == 0) && (length >= 2)) {
	return NewDiskObject(cell, argc-3, argv+3);
    }
    else if ((c == 's') && (strncmp(argv[2], "sphere", length) == 0) && (length >= 2)) {
	return NewSphereObject(cell, argc-3, argv+3);
    }
    else {
	Tcl_AppendResult(interp, "wrong object type \"", argv[2], "\": must be cylinder, disk, poly, or sphere.", (char *) NULL);
	return TCL_ERROR;
    }
}

/*
 *--------------------------------------------------------------
 *
 * NewPolyObject:
 *
 * create a new object of type 'poly'.
 *
 *--------------------------------------------------------------
 */

static int
NewPolyObject(cell, argc, argv)
     Cell *cell;
     int argc;
     char **argv;
{
    int new;
    Object *obj;
    Tcl_HashEntry *me;
    Tcl_Interp *interp = cell->interp;

    /* check for duplicate label */
    me = Tcl_CreateHashEntry(&cell->hash, argv[0], &new);
    if (!new) {
	Tcl_AppendResult(interp, "another object already has label \"", argv[0], "\".", (char *) NULL);
	return TCL_ERROR;
    }

    /* allocate memory for new object */
    if ((obj = (Object *) malloc(sizeof(Object))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }

    /* initialize object */
    obj->type = SM_TYPE_POLYGON;
    obj->next = NULL;
    obj->cell = cell;
    obj->child = NULL;
    obj->sibling = NULL;
    obj->quadric = NULL;
    obj->v = NULL;
    obj->s = NULL;
    obj->nv = obj->ns = 0;
    obj->e = me;
    obj->ns = obj->nv = 0;
    obj->color = NULL;
    obj->bfcolor = NULL;
    obj->parent = NULL;
    obj->normals = NULL;
    obj->invalid = 0;
    obj->updateFlag = SM_OBJECT_UPDATE_NORMAL;

    if ((obj->label = (char *) malloc(strlen(argv[0]) + 1)) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }
    (void) strcpy(obj->label, argv[0]);

    /* parse options */
    if (ConfigureObject(cell, obj, argc-1, argv+1, 0) != TCL_OK) {
	FreeObject(obj);
	return TCL_ERROR;
    }

    /* insert object into linked list and hashtable */
    Tcl_SetHashValue(me, (ClientData) obj);
    obj->next = cell->objects;
    cell->objects = obj;
    Tcl_AppendResult(interp, obj->label, (char *) NULL);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * NewCylinderObject:
 *
 * create a new object of type 'cylinder'.
 *
 *--------------------------------------------------------------
 */

static int
NewCylinderObject(cell, argc, argv)
     Cell *cell;
     int argc;
     char **argv;
{
    int new;
    Object *obj;
    Tcl_HashEntry *me;
    Tcl_Interp *interp = cell->interp;

    /* check for duplicate label */
    me = Tcl_CreateHashEntry(&cell->hash, argv[0], &new);
    if (!new) {
	Tcl_AppendResult(interp, "another object already has label \"", argv[0], "\".", (char *) NULL);
	return TCL_ERROR;
    }

    /* allocate memory for new object */
    if ((obj = (Object *) malloc(sizeof(Object))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }

    /* initialize object */
    obj->type = SM_TYPE_CYLINDER;
    obj->next = NULL;
    obj->cell = cell;
    obj->child = NULL;
    obj->sibling = NULL;
    obj->v = NULL;
    obj->s = NULL;
    obj->nv = obj->ns = 0;
    obj->e = me;
    obj->ns = obj->nv = 0;
    obj->color = NULL;
    obj->bfcolor = NULL;
    obj->parent = NULL;
    obj->normals = NULL;
    obj->invalid = 0;
    obj->updateFlag = 0;
    obj->quadric = NULL;

    if ((obj->label = (char *) malloc(strlen(argv[0]) + 1)) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }
    (void) strcpy(obj->label, argv[0]);
    obj->quadric = gluNewQuadric();

    /* parse options */
    if (ConfigureObject(cell, obj, argc-1, argv+1, 0) != TCL_OK) {
	FreeObject(obj);
	return TCL_ERROR;
    }

    /* insert object into linked list and hashtable */
    Tcl_SetHashValue(me, (ClientData) obj);
    obj->next = cell->objects;
    cell->objects = obj;
    Tcl_AppendResult(interp, obj->label, (char *) NULL);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * NewDiskObject:
 *
 * create a new object of type 'disk'.
 *
 *--------------------------------------------------------------
 */

static int
NewDiskObject(cell, argc, argv)
     Cell *cell;
     int argc;
     char **argv;
{
    int new;
    Object *obj;
    Tcl_HashEntry *me;
    Tcl_Interp *interp = cell->interp;

    /* check for duplicate label */
    me = Tcl_CreateHashEntry(&cell->hash, argv[0], &new);
    if (!new) {
	Tcl_AppendResult(interp, "another object already has label \"", argv[0], "\".", (char *) NULL);
	return TCL_ERROR;
    }

    /* allocate memory for new object */
    if ((obj = (Object *) malloc(sizeof(Object))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }

    /* initialize object */
    obj->type = SM_TYPE_DISK;
    obj->next = NULL;
    obj->cell = cell;
    obj->child = NULL;
    obj->sibling = NULL;
    obj->v = NULL;
    obj->s = NULL;
    obj->nv = obj->ns = 0;
    obj->e = me;
    obj->ns = obj->nv = 0;
    obj->color = NULL;
    obj->bfcolor = NULL;
    obj->parent = NULL;
    obj->normals = NULL;
    obj->invalid = 0;
    obj->updateFlag = 0;
    obj->quadric = NULL;

    if ((obj->label = (char *) malloc(strlen(argv[0]) + 1)) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }
    (void) strcpy(obj->label, argv[0]);
    obj->quadric = gluNewQuadric();

    /* parse options */
    if (ConfigureObject(cell, obj, argc-1, argv+1, 0) != TCL_OK) {
	FreeObject(obj);
	return TCL_ERROR;
    }

    /* insert object into linked list and hashtable */
    Tcl_SetHashValue(me, (ClientData) obj);
    obj->next = cell->objects;
    cell->objects = obj;
    Tcl_AppendResult(interp, obj->label, (char *) NULL);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * NewSphereObject:
 *
 * create a new object of type 'sphere'.
 *
 *--------------------------------------------------------------
 */

static int
NewSphereObject(cell, argc, argv)
     Cell *cell;
     int argc;
     char **argv;
{
    int new;
    Object *obj;
    Tcl_HashEntry *me;
    Tcl_Interp *interp = cell->interp;

    /* check for duplicate label */
    me = Tcl_CreateHashEntry(&cell->hash, argv[0], &new);
    if (!new) {
	Tcl_AppendResult(interp, "another object already has label \"", argv[0], "\".", (char *) NULL);
	return TCL_ERROR;
    }

    /* allocate memory for new object */
    if ((obj = (Object *) malloc(sizeof(Object))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }

    /* initialize object */
    obj->type = SM_TYPE_SPHERE;
    obj->next = NULL;
    obj->cell = cell;
    obj->child = NULL;
    obj->sibling = NULL;
    obj->v = NULL;
    obj->s = NULL;
    obj->nv = obj->ns = 0;
    obj->e = me;
    obj->ns = obj->nv = 0;
    obj->color = NULL;
    obj->bfcolor = NULL;
    obj->parent = NULL;
    obj->normals = NULL;
    obj->invalid = 0;
    obj->updateFlag = 0;
    obj->quadric = NULL;

    if ((obj->label = (char *) malloc(strlen(argv[0]) + 1)) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for new object.", (char *) NULL);
	Tcl_DeleteHashEntry(me);
	return TCL_ERROR;
    }
    (void) strcpy(obj->label, argv[0]);
    obj->quadric = gluNewQuadric();

    /* parse options */
    if (ConfigureObject(cell, obj, argc-1, argv+1, 0) != TCL_OK) {
	FreeObject(obj);
	return TCL_ERROR;
    }

    /* insert object into linked list and hashtable */
    Tcl_SetHashValue(me, (ClientData) obj);
    obj->next = cell->objects;
    cell->objects = obj;
    Tcl_AppendResult(interp, obj->label, (char *) NULL);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ConfigureObject:
 *
 * configure various aspects of an object.
 *
 *--------------------------------------------------------------
 */

static int
ConfigureObject(cell, obj, argc, argv, flags)
     Cell *cell;
     Object *obj;
     int argc;
     char **argv;
     int flags;
{
    DOUBLE r, g, b;
    DOUBLE br, bg, bb;
    int updateColor = 0;
    int updateBFColor = 0;
    Tk_ConfigSpec *specs;
    Tk_Window tkwin = Tk_MainWindow(cell->interp);
    XColor *color = obj->color;
    XColor *bfcolor = obj->bfcolor;
    
    obj->updateFlag = 0;

    SmSetClientData((ClientData) GetCellHeader(cell->interp));
    switch(obj->type) {
      case SM_TYPE_POLYGON:
	specs = objectPolygonConfigSpecs;
	break;

      case SM_TYPE_DISK:
	specs = objectDiskConfigSpecs;
	break;

      case SM_TYPE_CYLINDER:
	specs = objectCylinderConfigSpecs;
	break;

      case SM_TYPE_SPHERE:
	specs = objectSphereConfigSpecs;
	break;
    }

    if (Tk_ConfigureWidget(cell->interp, tkwin, specs, argc, argv, (char *) obj, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    if (obj->updateFlag & SM_OBJECT_UPDATE_NORMAL) {
	if (ComputeVertexNormals(obj) != TCL_OK) {
	    if (obj->normals) free((void *) obj->normals);
	    obj->normals = NULL;
	    return TCL_ERROR;
	}
    }
    if (obj->updateFlag & SM_OBJECT_UPDATE_POSITION) {
	obj->invalid = ComputeModelTransform(obj, cell->epsilon);
    }

    if (obj->color != color) {
	color = obj->color;
	updateColor = 1;
    }
    r = color->red / 65535.0;
    g = color->green / 65535.0;
    b = color->blue / 65535.0;

    if (updateColor || (obj->updateFlag & SM_OBJECT_UPDATE_AMBIENT)) {
	obj->ambientRGBA[0] = obj->ambient[0] * r;
	obj->ambientRGBA[1] = obj->ambient[1] * g;
	obj->ambientRGBA[2] = obj->ambient[2] * b;
	obj->ambientRGBA[3] = 1.0;
    }

    if (updateColor || (obj->updateFlag & SM_OBJECT_UPDATE_DIFFUSE)) {
	obj->diffuseRGBA[0] = obj->diffuse[0] * r;
	obj->diffuseRGBA[1] = obj->diffuse[1] * g;
	obj->diffuseRGBA[2] = obj->diffuse[2] * b;
	obj->diffuseRGBA[3] = 1.0;
    }

    if (updateColor || (obj->updateFlag & SM_OBJECT_UPDATE_EMISSION)) {
	obj->emissionRGBA[0] = obj->emission[0] * r;
	obj->emissionRGBA[1] = obj->emission[1] * g;
	obj->emissionRGBA[2] = obj->emission[2] * b;
	obj->emissionRGBA[3] = 1.0;
    }

    if (updateColor || (obj->updateFlag & SM_OBJECT_UPDATE_SPECULAR)) {
	obj->specularRGBA[0] = obj->specular.color->red / 65535.0;
	obj->specularRGBA[1] = obj->specular.color->green / 65535.0;
	obj->specularRGBA[2] = obj->specular.color->blue / 65535.0;
	obj->specularRGBA[3] = 1.0;
    }

    if (obj->bfcolor != bfcolor) {
	bfcolor = obj->bfcolor;
	updateBFColor = 1;
    }
    br = bfcolor->red / 65535.0;
    bg = bfcolor->green / 65535.0;
    bb = bfcolor->blue / 65535.0;

    if (updateBFColor || (obj->updateFlag & SM_OBJECT_UPDATE_BF_AMBIENT)) {
	obj->bfambientRGBA[0] = obj->bfambient[0] * br;
	obj->bfambientRGBA[1] = obj->bfambient[1] * bg;
	obj->bfambientRGBA[2] = obj->bfambient[2] * bb;
	obj->bfambientRGBA[3] = 1.0;
    }

    if (updateBFColor || (obj->updateFlag & SM_OBJECT_UPDATE_BF_DIFFUSE)) {
	obj->bfdiffuseRGBA[0] = obj->bfdiffuse[0] * br;
	obj->bfdiffuseRGBA[1] = obj->bfdiffuse[1] * bg;
	obj->bfdiffuseRGBA[2] = obj->bfdiffuse[2] * bb;
	obj->bfdiffuseRGBA[3] = 1.0;
    }

    if (updateBFColor || (obj->updateFlag & SM_OBJECT_UPDATE_BF_EMISSION)) {
	obj->bfemissionRGBA[0] = obj->bfemission[0] * br;
	obj->bfemissionRGBA[1] = obj->bfemission[1] * bg;
	obj->bfemissionRGBA[2] = obj->bfemission[2] * bb;
	obj->bfemissionRGBA[3] = 1.0;
    }

    if (updateBFColor || (obj->updateFlag & SM_OBJECT_UPDATE_BF_SPECULAR)) {
	obj->bfspecularRGBA[0] = obj->bfspecular.color->red / 65535.0;
	obj->bfspecularRGBA[1] = obj->bfspecular.color->green / 65535.0;
	obj->bfspecularRGBA[2] = obj->bfspecular.color->blue / 65535.0;
	obj->bfspecularRGBA[3] = 1.0;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ComputeModelTransform:
 *
 * compute the model transform matrix for an object.
 *
 *--------------------------------------------------------------
 */

static int
ComputeModelTransform(obj, epsilon)
     Object *obj;
     double epsilon;
{
    int i, j;
    Surface *s;
    DOUBLE x[3], y[3], z[3];

    /* check fwd and up vectors */
    if (SmNormalizeVector3D(obj->fwd, z, epsilon)) return 1;

    if (SmNormalizeVector3D(obj->up, y, epsilon)) return 1;

    SmCrossVector3D(y, z, x);
    if (SmNormalizeVector3D(x, x, epsilon)) return 1;

    /* make sure vertex indices are within bounds */
    for (i = 0, s = obj->s; i < obj->ns; s++, i++) {
	for (j = 0; j < s->vcnt; j++) {
	    if (s->index[j] < 0 || s->index[j] >= obj->nv) {
		return 1;
	    }
	}
    }

    /* set up model transform matrix */
    obj->modelT[0] = x[0];
    obj->modelT[1] = x[1];
    obj->modelT[2] = x[2];
    obj->modelT[3] = 0;

    obj->modelT[4] = y[0];
    obj->modelT[5] = y[1];
    obj->modelT[6] = y[2];
    obj->modelT[7] = 0;

    obj->modelT[8] = z[0];
    obj->modelT[9] = z[1];
    obj->modelT[10] = z[2];
    obj->modelT[11] = 0;

    obj->modelT[12] = obj->pos[0];
    obj->modelT[13] = obj->pos[1];
    obj->modelT[14] = obj->pos[2];
    obj->modelT[15] = 1;

    /* passed all tests */
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ComputeVertexNormals:
 *
 * compute vertex normals.
 *
 *--------------------------------------------------------------
 */

static int
ComputeVertexNormals(obj)
     Object *obj;
{
    int i, j;
    int argc;
    char **argv;
    Vertex *v;
    Surface *s;
    Cell *cell = obj->cell;
    Tcl_Interp *interp = cell->interp;

    /* reset vertex normals */
    for (s = obj->s, i = 0; i < obj->ns; s++, i++) {
	s->normal = SM_VERTEX_NORMAL_DEFAULT;
    }

    /* parse user-specified normals */
    if (obj->normals) {
	if (Tcl_SplitList(interp, obj->normals, &argc, &argv) != TCL_OK) {
	    free((void *) argv);
	    return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
	    if (ParseVertexNormals(interp, obj, argv[i]) != TCL_OK) {
		free((void *) argv);
		return TCL_ERROR;
	    }
	}
        free((void *) argv);
    }

    /* reset vertex normal to zero */
    for (v = obj->v, i = 0; i < obj->nv; v++, i++) {
	v->lvn.v[0] = v->lvn.v[1] = v->lvn.v[2] = 0;
    }

    /* sum surface normals */
    for (s = obj->s, i = 0; i < obj->ns; s++, i++) {
	SmComputeSurfaceNormal(obj->v, s, s->ln, cell->epsilon);
	for (j = 0; j < s->vcnt - 1; j++) {
	    v = &obj->v[s->index[j]];
	    v->lvn.v[0] += s->ln[0];
	    v->lvn.v[1] += s->ln[1];
	    v->lvn.v[2] += s->ln[2];
	}
    }

    /* normalize vertex normals */
    for (v = obj->v, i = 0; i < obj->nv; v++, i++) {
	SmFastNormalizeVector3D(v->lvn.v, v->lvn.v);
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ParseVertexNormals:
 *
 * parse vertex normals for a single surface.
 *
 *--------------------------------------------------------------
 */

static int
ParseVertexNormals(interp, obj, value)
     Tcl_Interp *interp;
     Object *obj;
     char *value;
{
    int i, argc, nc, index;
    char **argv, **coords;
    Normal3D *n;
    Surface *s;
    double v[3];
    Cell *cell = obj->cell;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc < 1) {
	free((void *) argv);
	Tcl_AppendResult(interp, "bad format: ", value, (char *) NULL);
	return TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[0], &index) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }

    if (index < 1 || index > obj->ns) {
	free((void *) argv);
	Tcl_AppendResult(interp, "bad surface index: ", argv[0], (char *) NULL);
	return TCL_ERROR;
    }
    s = &obj->s[index - 1];

    /* number of normals must equal number of vertices minus one */
    if (argc != 1 && argc != s->vcnt) {
	free((void *) argv);
	Tcl_AppendResult(interp, "bad format: number of vertex normals must equal either 0 or number of vertices.",
			 (char *) NULL);
	return TCL_ERROR;
    }

    for (i = 1, n = s->lvn; i < argc; i++, n++) {
	if (Tcl_SplitList(interp, argv[i], &nc, &coords) != TCL_OK) {
	    free((void *) coords);
	    free((void *) argv);
	    return TCL_ERROR;
	}
	if (nc != 3) {
	    Tcl_AppendResult(interp, "bad vector format: 3 components required.", (char *) NULL);
	    free((void *) coords);
	    free((void *) argv);
	    return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, coords[0], &v[0]) != TCL_OK ||
	    Tcl_GetDouble(interp, coords[1], &v[1]) != TCL_OK ||
	    Tcl_GetDouble(interp, coords[2], &v[2]) != TCL_OK) {
	    free((void *) coords);
	    free((void *) argv);
	    return TCL_ERROR;
	}
	n->v[0] = v[0];
	n->v[1] = v[1];
	n->v[2] = v[2];

	if (SmNormalizeVector3D(n->v, n->v, cell->epsilon)) {
	    Tcl_AppendResult(interp, "illegal vertex normal: ", argv[i], (char *) NULL);
	    free((void *) coords);
	    free((void *) argv);
	    return TCL_ERROR;
	}
	free((void *) coords);
    }
    s->normal = (argc == 1) ? SM_VERTEX_NORMAL_DEFAULT : SM_VERTEX_NORMAL_USER;
    free((void *) argv);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SmComputeSurfaceNormal:
 *
 * compute the surface normal vector from vertex positions.
 *
 *--------------------------------------------------------------
 */

void
SmComputeSurfaceNormal(v, s, n, epsilon)
     Vertex *v;
     Surface *s;
     DOUBLE n[3];
     double epsilon;
{
    Vertex *v0, *v1, *v2;
    DOUBLE a[3], b[3];

    v0 = &v[s->index[0]];
    v1 = &v[s->index[1]];
    v2 = &v[s->index[2]];

    a[0] = v0->lv[0] - v1->lv[0];
    a[1] = v0->lv[1] - v1->lv[1];
    a[2] = v0->lv[2] - v1->lv[2];

    b[0] = v1->lv[0] - v2->lv[0];
    b[1] = v1->lv[1] - v2->lv[1];
    b[2] = v1->lv[2] - v2->lv[2];
	
    SmCrossVector3D(a, b, n);
    if (SmNormalizeVector3D(n, n, epsilon)) {
	n[0] = n[1] = n[2] = 0;
    }
}
